


var Analyzer = (function _Analyzer() {
    
    const fs = require('fs');
    const path = require("path");
    const Parse = require('./parse_tree.js');
    
    var s = {
        pos : ["NOUN", "VERB", "ADJF", "ADJS", "COMP",
            "INFN", "PRTF", "PRTS", "GRND", "NUMR", "ADVB",
            "NPRO", "PREP", "PRED", "CONJ", "PRCL", "INTJ", "QUES"],
        valency : ["tran", "intr"],
        numb : ["sing", "plur"],
        gend : ["musc", "femn", "neut"],
        per : ["1per", "2per", "3per"],
        tense: ["past", "pres", "futr"],
        cases : ["nomn", "gent", "datv", "accs", "acc2", "gen1", "gen2", "ablt", "loct", "voct", "loc1", "loc2"],
        dict: JSON.parse(fs.readFileSync(path.dirname(require.resolve("syntax_analyzer/package.json")) + '/dict/dict.json')),
        grammar: JSON.parse(fs.readFileSync(path.dirname(require.resolve("syntax_analyzer/package.json"))+'/grammar.json'))
    }, Az;
    
    String.prototype.replaceAll = function (search, replacement) {
        var target = this;
        return target.split(search).join(replacement);
    };
    
    function parse(sent) {
        var result = [];
        
        // delete all punctuation marks
        sent = sent.replaceAll(',', '');
        sent = sent.replaceAll('!', '');
        sent = sent.replaceAll('?', '');
        sent = sent.replaceAll('.', '');
        sent = sent.replaceAll('"', '');
        sent = sent.replaceAll(':', '');
        
        // find complex pharses (that are consits of two or more words)
        for (var i =0; i<s.dict['conj'].length; i++) {
            if(sent.indexOf(s.dict['conj'][i])!==-1) {
                sent = sent.replaceAll(s.dict['conj'][i], '$[conj]'+s.dict['conj'][i]+'$')
            }
        }
        
        for (var i =0; i<s.dict['pred'].length; i++) {
            if(sent.indexOf(s.dict['pred'][i])!==-1) {
                sent = sent.replaceAll(s.dict['pred'][i], '$[pred]'+s.dict['pred'][i]+'$')
            }
        }
        
        sent = sent.split('$');
        
        for(var i = 0; i < sent.length; i++){
            sent[i] = sent[i].trim();
            if(sent[i] == "") {
                sent.splice(i,1);
                i--; // Prevent skipping an item
            }
        }
        
        for(var i=0;i< sent.length; i++) {
            if(sent[i].length!=0) {
                if(sent[i][0]!='[') {
                    sent[i] = sent[i].split(" ");
                }
            }
        }
        
        var new_sent = [];
        
        for(var i=0;i< sent.length; i++) {
            
            if(typeof sent[i] == "object") {
                for(var j=0;j<sent[i].length;j++) {
                    new_sent.push(sent[i][j]);
                }
            } else {
                new_sent.push(sent[i]);
            }
        }
        
        // find grammemes to each word
        for(var i in new_sent) {
            
            if(new_sent[i][0]=="[") {
                result.push([new_sent[i].split("]")[1],tag(new_sent[i])]);
            } else {
                result.push([new_sent[i],tag(new_sent[i])]);
            }
        }

        return _parse(result);
    }
    
    // itertool like product func
    function product(arr) {
        
        return arr.reduce(function tl (accumulator, value) {
            var tmp = [];
            accumulator.forEach(function (a0) {
                value.forEach(function (a1) {
                    tmp.push(a0.concat(a1));
                });
            });
            return tmp;
        }, [[]]);
        
    }
    
    function create_parse_tree(sent) {
        
        var tree = new Parse.Tree(s.grammar);

        tree.build(sent);
        tree.reduce();

        if (!tree.create_root([tree])) {
            return false;
        } else {
            return tree;
        }
    }
    
    
    // function to split the sentence into clauses 
    
    function split_sentence(sent) {
        
        var temp = sent.slice();
        var subtrees = [];
        for (var i=0; i< temp.length; i++) {
            
            // if conj - try to use it as delimiter of clauses
            if (temp[i][1] == "CONJ") {
                if(create_parse_tree(temp.slice(0,i))){
                    subtrees.push(create_parse_tree(temp.slice(0,i)));
                    temp.splice(0,i+1);
                }
            }
            
            // if nominative - try to use it as delimiter of clauses
            if (i<temp.length && temp[i][1] == "NP[case='nomn']") {
                if(create_parse_tree(temp.slice(0,i)) && create_parse_tree(temp.slice(i,temp.length))) {
                    subtrees.push(create_parse_tree(temp.slice(0,i)));
                    temp.splice(0,i);
                }
            }
        }

        var pt = create_parse_tree(temp);
        
        if(!pt) {
            return false;
        } else {
            subtrees.push(pt);
            tree = new Parse.Tree(s.grammar);
            tree.create_root(subtrees);
            return tree;
        }
        
    }
    
    function build_trees(sent) {
        
        var variants = [];
        var trees = [];
        var result=[];
        
        var p = product(sent);
        
        for(var i = 0; i< p.length; i++) {
            variants[i] = [];
            while(p[i].length!==0) {
                variants[i].push(p[i].splice(0,3));
            }
        }
        
        for(var i= 0; i< variants.length;i++) {

            var ss = split_sentence(variants[i]);
            trees.push(ss);
        }
        
        for(var i=0;i<trees.length;i++) {
            if(trees[i]) {
                result.push(trees[i]); 
            }
        }
        
        return result.slice(0,5);
        
    }
    // find lefthand part of production   
    
    function find_lhs(rhs) {
        
        var rule = general_form(rhs);
        
        // here iterate up to stem
        
        function iterate(r) {
            for(var key in s.grammar) {
                if(s.grammar[key].indexOf(r)!==-1) {
                    return iterate(key);
                }
            }
            return r;
        }
        
        rule = iterate(rule);
        
        return rule;
    }
    
    function general_form(form) {
        
        var result = [];
        
        result[0] = form[0];
        result[1]="?numb";
        result[2]="?per";
        result[3]="?gend";
        
        if (!form[4] ){
            result[4] = "null";
        } else {
            result[4] = form[4];
        }
        
        
        if (!form[5]) {
            result[5] = "null";
        } else {
            result[5] = form[5];
        }
        
        
        if (!form[6]) {
            result[6] = "null";
        } else {
            result[6] = '?tense';
        }
        
        return "["+result.toString().replaceAll('"','')+"]";
    }
    
    
    function _parse(sent) {
        
        result = [];
        var variants;
        
        for(var i = 0; i<sent.length;i++) {
            
            variants = [];
            
            for(var x = 0; x<sent[i][1].length; x++) {
                variants.push([sent[i][0], find_lhs(sent[i][1][x]), sent[i][1][x]]);
            }
            
            result.push(variants.slice(0,5));
        }

        trees = build_trees(result);
        
        return trees;
    }
    
    function tag(word) {
        
        var result = [];
        var tags_set = [null,null,null,null,null,null,null];
        var grammeme, tags;
        
        // if word is complex pharse (predicative of conjugation)
        
        if (word.substring(1,5) == 'pred') {
            var tags = ["PRED",null,null,null,null,null,null];
            return [tags];
        }
        
        if (word.substring(1,5) == 'conj') {
            var tags = ["CONJ",null,null,null,null,null,null];
            return [tags];
        }
        
        // if word is single
        
        var parses = Az.Morph(word);
        
        for(var i =0; i<parses.length;i++) {
            tags = parses[i].tag.stat.concat(parses[i].tag.flex);
            for(var x in tags) {
                grammeme = tags[x];
                
                if(s.numb.indexOf(grammeme)!==-1) {
                    tags_set[1] = grammeme;
                }
                if(s.per.indexOf(grammeme)!==-1) {
                    tags_set[2] = grammeme;
                }
                if(s.gend.indexOf(grammeme)!==-1) {
                    tags_set[3] = grammeme;
                }
                if(s.cases.indexOf(grammeme)!==-1) {
                    tags_set[4] = grammeme;
                }
                if(s.valency.indexOf(grammeme)!==-1) {
                    tags_set[5] = grammeme;
                }
                if(s.tense.indexOf(grammeme)!==-1) {
                    tags_set[6] = grammeme;
                }
                
                
                if (s.pos.indexOf(grammeme)!==-1 && !tags_set[0]) {
                    tags_set[0] = grammeme;
                }
                
                // if word is anaphora 
                if (grammeme == "Anph") {
                    tags_set[0] = "NPRO";
                }
                
                // if word is predicative
                if (grammeme == "Prdx") {
                    tags_set[0] = "PRED"
                }
            }
            result.push(tags_set);
            tags_set = [null,null,null,null,null,null,null];
        }
        
        return result;
    }
    
    function _initialize(AZ) {
        
        if(!AZ) {
            return "load Az module first.";
        }
        Az = AZ;
        
    }
    
    return {
        initialize: _initialize,
        parse: parse
    };
})();

module.exports = Analyzer;