
var Parse = (function _Parse() {
    
    var grammarKeys;
    
    var Root = function() {
        var s = {};
        s.sentences = [];
        return s;
    };
    
    var Node = function(word = null, tag = null, grammemes = null, leaf = null) {
        
        function equal(arr1, arr2) {
            for(var i = 0; i<arr1.length; i++) {
                if(arr1[i] !== arr2[1]) {
                    return false;
                }
            }
            return true;
        }
        
        var s = {};
        s.word = word;
        s.tag = tag;
        s.grammemes = grammemes;
        s.grammemes_simple = null;
        
        if (grammemes) {
            s.grammemes_simple = [];
            for(var i = 0; i<s.grammemes.length; i++) {
                if(s.grammemes[i]!==null) {
                    s.grammemes_simple.push(s.grammemes[i]);
                }
            }
        }
        
        s.leaf = leaf;
        s.debug = function() {
            console.log(s.word + " "+s.tag + " "+(s.l ? s.l.tag : null) + " " + (s.r ? s.r.tag : null) + " "+s.p);
        };
        s.isEqual = function(node) {
            return (s.tag == node.tag && s.word == node.word && equal(s.grammemes,node.grammemes) && equal(s.grammemes_simple,node.grammemes_simple))
        };
        s.l = null;
        s.r = null;
        s.p = null;
        
        return s;
    };
    
    var Tree = function(grammar) {
        var s = {};
        s.grammar = grammar;
        s.root = null;
        s.nodes = [];
        
        if(!grammarKeys) {
            grammarKeys = [];
            for(var key in s.grammar) {
                grammarKeys = grammarKeys.concat(s.grammar[key]);
            }
        }
        
        function _range(start, stop) {
            
            if (start >= stop) {
                return [];
            }
            
            var result = [];
            for (var i = start; i < stop; i++) {
                result.push(i);
            }
            
            return result;
        };
        
        
        function get_parent_rule(r) {
            
            for(var key in s.grammar) {
                if(s.grammar[key].indexOf(r)!==-1) {
                    
                    return key;
                }
            }
            return r;
        }
        
        function find_rule(node1, node2) {
            
            var rule1 = node1.tag + ' ' + node2.tag;
            var rule2 = node2.tag + ' ' + node1.tag;
            
            if(grammarKeys.indexOf(rule1)!==-1) {
                return rule1;
            }
            if(grammarKeys.indexOf(rule2)!==-1) {
                return rule2;
            }
            
            return null;
        }
        
        function agreement(node_left, node_right) {
            
            if (node_left.grammemes && node_right.grammemes) {
                
                var numb1 = node_left.grammemes[1];
                var numb2 = node_right.grammemes[1];
                
                var per1 = node_left.grammemes[2];
                var per2 = node_right.grammemes[2]; 
                
                var gend1 = node_left.grammemes[3]; 
                var gend2 = node_right.grammemes[3];
                
                if (numb1 && numb2) {
                    if (numb1 != numb2) {
                        return false;
                    } 
                }
                
                if (per1 && per2) {
                    if (per1 != per2) {
                        return false;
                    } 
                }
                
                if (gend1 && gend2) {
                    if (gend1 != gend2) {
                        return false;
                    } 
                }
                
            }
            return true;
        }
        
        function unite_nodes(node_left, node_right, parent_tag) {
            
            var parent_node = new Node(null, parent_tag);
            
            // parent takes grammemes of main word (NOUN)
            
            if (parent_tag.substr(0,2) == 'NP') {
                
                if (node_left.grammemes) {
                    if (node_left.tag != "NP[case='gent']") {
                        parent_node.grammemes = node_left.grammemes;
                    }
                    
                }
                
                if (node_right.grammemes) {
                    if (node_right.tag != "NP[case='gent']") {
                        parent_node.grammemes = node_right.grammemes;
                    }
                    
                }
            }
            
            // parent takes grammemes of main word (VERB)
            
            if (parent_tag.substr(0,2) == 'VP') {
                
                
                if (node_left.grammemes) {
                    if (node_left.grammemes[0] == 'VERB' || node_left.grammemes[0] == "INFN") {
                        parent_node.grammemes = node_left.grammemes;
                    }
                    
                    if (node_right.grammemes) {
                        if (node_right.grammemes[0] == 'VERB' || node_right.grammemes[0] == 'INFN' ) {
                            parent_node.grammemes = node_right.grammemes;
                        }
                    }
                }
            }
            
            // parent takes grammemes of main word (NOUN)
            
            if (parent_tag.substr(0,2) == 'PP') {
                if (node_left.grammemes) {
                    if (node_left.grammemes[0] == 'NOUN' || node_left.grammemes[0] == 'NPRO') {
                        parent_node.grammemes = node_left.grammemes;
                    }
                    
                    if (node_right.grammemes) {
                        if (node_right.grammemes[0] == 'NOUN' || node_right.grammemes[0] == 'NPRO' ) {
                            parent_node.grammemes = node_right.grammemes;
                        }
                    }
                }
            }
            
            
            node_left.p = parent_node;
            node_right.p = parent_node;
            parent_node.l = node_left;		
            parent_node.r = node_right;
            
            var position = s.nodes.indexOf(node_left);
            s.nodes.splice(position, 0, parent_node);
            s.nodes.splice(s.nodes.indexOf(node_left),1);
            s.nodes.splice(s.nodes.indexOf(node_right),1);
        }
        
        function reduce_VP() {
            
            for (var i=0; i< s.nodes.length;i++) {
                var node1 = s.nodes[i];
                for(var j=0; j<s.nodes.length; j++) {
                    var node2 = s.nodes[j];
                    
                    if (node1 != node2 && i<s.nodes.length && j<s.nodes.length) {
                        
                        var rule = find_rule(s.nodes[i], s.nodes[j]);
                        var rr = get_parent_rule(rule);
                        
                        if (rule && (rr.substr(0,2) == 'VP' || rr.substr(0,4) == 'PRED')) {
                            
                            unite_nodes(s.nodes[i], s.nodes[j], rr);
                            reduce_VP();
                        }
                    }
                }
            }
        }
        
        function reduce_ADJ() {
            for (var i=0; i< s.nodes.length;i++) {
                
                if (s.nodes[i].tag.substr(0,3) == "ADJ" || s.nodes[i].tag.substr(0,3) == "PRT") {
                    var rule = null, rr = null;
                    
                    if ( i > 0 ) {
                        rule = find_rule(s.nodes[i-1], s.nodes[i]);
                        rr = get_parent_rule(rule);
                    }
                    
                    if (rule && (rr.substr(0,3) == 'ADJ' || rr.substr(0,3) == 'PRT')) {
                        unite_nodes(s.nodes[i-1], s.nodes[i], rr);
                        reduce_ADJ();
                    }
                    
                    rule = null, rr = null;
                    
                    if (i+1 < s.nodes.length ) {
                        j = i + 1;
                        if (s.nodes[i+1].tag == "CONJ" && i+2 < s.nodes.length) {
                            if (s.nodes[i+2].tag.substr(0,3) == "ADJ" && s.nodes[i].tag.substr(0,3) == "ADJ") {
                                j = i+2;
                            } else {
                                rule2 = null;
                            }
                            
                            if (s.nodes[i+2].tag.substr(0,3) == "PRT" && s.nodes[i].tag.substr(0,3) == "PRT") {
                                j = i+2;
                            } else {
                                rule2 = null;
                            }
                        }
                        
                        rule = find_rule(s.nodes[i], s.nodes[j]);
                        rr =  get_parent_rule(rule);
                    }
                    
                    if (rule && (rr.substr(0,3) == 'ADJ' || rr.substr(0,3) == 'PRT')) {
                        unite_nodes(s.nodes[i], s.nodes[j], rr);
                        reduce_ADJ();
                    }
                }
            }
        }
        
        function reduce_NP() {
            
            var rule, rr;
            
            for (var i=0; i< s.nodes.length;i++) {
                
                if (s.nodes[i].tag.substr(0,2) == "NP") {
                    rule = null, rr = null;
                    
                    if ( i > 0 ) {
                        rule = find_rule(s.nodes[i-1], s.nodes[i]);
                        rr = get_parent_rule(rule);
                    }
                    
                    if (rr && rr.substr(0,2) == 'NP') {
                        unite_nodes(s.nodes[i-1], s.nodes[i], rr);
                        reduce_NP();
                    }
                    
                    rule = null, rr = null;
                    
                    if ( i+1 < s.nodes.length ) {
                        j = i + 1;
                        if (s.nodes[i+1].tag == "CONJ" && i+2 < s.nodes.length) {
                            if (s.nodes[i+2].tag.substr(0,2) == "NP" && s.nodes[i].tag.substr(0,2) == "NP") {
                                j = i+2;
                            } else {
                                rule2 = null;
                            }
                        }
                        
                        rule = find_rule(s.nodes[i], s.nodes[j]);
                        rr = get_parent_rule(rule);
                    }
                    
                    if (rr && rr.substr(0,2) == 'NP') {
                        
                        unite_nodes(s.nodes[i], s.nodes[j], rr);
                        reduce_NP();
                    }
                }
            }
        }
        
        function reduce_PP(self) {
            for (var i = 0; i < s.nodes.length;i++) {
                if (s.nodes[i].tag == "PREP" ||	s.nodes[i].tag == "PP") {
                    var rule = null, rr = null;
                    
                    if ( i > 0 ) {
                        rule = find_rule(s.nodes[i-1], s.nodes[i]);
                        rr = get_parent_rule(rule);
                    }
                    
                    if (rr && rr.substr(0,2) == 'PP') {
                        unite_nodes(s.nodes[i-1], s.nodes[i], rr);
                        reduce_PP();
                    }
                    
                    rule = null, rr = null;
                    
                    if ( i+1 < s.nodes.length ) {
                        
                        j = i + 1;
                        if (s.nodes[i+1].tag == "CONJ" && (i+2 < s.nodes.length)) {
                            if (self.nodes[i+2].tag.substr(0,2) == "PP") {
                                j = i+2;
                            } else {
                                rule2 = null;
                            }
                        }
                        
                        rule = find_rule(s.nodes[i], s.nodes[j]);
                        rr = get_parent_rule(rule);
                    }
                    
                    if (rr && rr.substr(0,2) == 'PP') {
                        unite_nodes(s.nodes[i], s.nodes[j], rr);
                        reduce_PP();   
                    }
                }
            }
        }
        function reduce_S() {
            
            for (var i=0; i< s.nodes.length;i++) {
                
                var range = _range(i+1, s.nodes.length);
                
                for (var j=0; j< range.length;j++) {
                    
                    if (i < s.nodes.length && range[j] < s.nodes.length) {
                        
                        var rule = find_rule(s.nodes[i], s.nodes[range[j]]);
                        var rr = get_parent_rule(rule);
                        
                        if (rule && rr.substr(0,2) == 'S' && agreement(s.nodes[i], s.nodes[range[j]])) {
                            
                            unite_nodes(s.nodes[i], s.nodes[range[j]], rr);
                            reduce_S();
                        }
                    }
                    
                }
            }
            for (var i=0; i< s.nodes.length; i++) {
                if (s.nodes[i].tag == "CONJ") {
                    s.nodes.splice(s.nodes.indexOf(s.nodes[i]),1);
                }
            }
            
        }
        
        function _display(node, space) {
            
            if (node) {
                
                if (node.tag == "S") {
                    console.log(node.tag);
                }
                
                if (node.tag != "S" && !node.leaf) {
                    console.log(" ".repeat(space) + node.tag);
                }
                
                if (node.leaf) {
                    console.log(" ".repeat(space * 2) + node.tag, '\n', " ".repeat(space * 3), node.word, node.grammemes_simple);
                }
                
                _display(node.l, space+2);
                _display(node.r, space+2);  
            }
            
        }
        
        function _extract(tag, node, result) {
            
            result = result || [];
            
            if(node) {
                if(!node.leaf) {
                    if((node.l.word && (node.l.tag == tag && node.l.leaf)) || (node.r.word && node.r.tag == tag && node.r.leaf)) {
                        if(node.l.word) {
                            result.push({
                                word: node.l.word,
                                grammemes: node.l.grammemes_simple
                            });
                        }
                        if(node.r.word) {
                            result.push({
                                word: node.r.word,
                                grammemes: node.r.grammemes_simple
                            });
                        }
                    } else {
                        result = _get(tag,node.r,result);
                        result = _get(tag,node.l,result);
                    }
                    
                } 
            }
            return result;
        }
        
        function get_sentence(node, sent) {
            if (node) {
                get_sentence(node.l, sent);
                if (node.word) {
                    sent.push([node.word, node.grammemes_simple]);
                    get_sentence(node.r, sent);
                    return sent;
                }
            }	
        }
        
        function debug() {
            console.log("TREE DEBUG ---------");
            for(var i = 0; i<s.nodes.length; i++) {
                s.nodes[i].debug();
            }
            console.log("---------");
        }
        
        return {
            
            build: function(sent) {
                
                for(var i= 0; i<sent.length;i++) {
                    var new_node = new Node(sent[i][0], sent[i][1], sent[i][2], true);
                    s.nodes.push(new_node);
                }
                
            },
            
            reduce: function() {
                
                reduce_ADJ();
                reduce_NP();
                reduce_PP();
                reduce_VP();
                reduce_S();
                
            },
            
            create_root: function(subtrees) {
                s.root = new Root();
                
		for(var i = 0; i<subtrees.length;i++) {
                    
                    var subtree = subtrees[i];
                    var nodes = subtree.nodes();
                    
                    for(var n = 0; n < nodes.length; n++) {
                        
                        var node = nodes[n];
                        
                        if (node.tag !== 'S' && node.tag.substr(0,2) !== 'VP') {
                            
                            return false;
                            
                        } else {
                            // complete sentence
                            if (node.tag === 'S') {
                                s.root.sentences.push(node);
                            }
                            
                            // incomplete sentence
                            if (node.tag.substr(0,2) === "VP") {
                                s.root.sentences.push(node);
                            }
                        }
                    }
                }
                
		if (s.root.sentences.length === 0) {
                    return false;
                } else {
                    return true;
                }
            },
            
            display: function() {
                
                for (var i = 0; i <s.root.sentences.length; i++) {
                    
                    _display(s.root.sentences[i],1);
                }
            },
            
            extract: function(tag) {
                // get nodes arr by tag
                // used for extracting a part of sentence by syntax
                var result = [];
                for(var i = 0; i < s.root.sentences.length; i++) {
                    result = _get(tag, s.root.sentences[i], result);
                }
                return result;
            },
            
            nodes: function() {
                return s.nodes;
            },
            debug: debug
        };
        
    };
    
    return {
        Node: Node,
        Root: Root,
        Tree: Tree
    };
})();

module.exports = Parse;